from .compat import viewitems
from ._common import BidirectionalMapping
from collections import Hashable

class frozenbidict(BidirectionalMapping, Hashable):
    """
    Immutable, hashable bidict type.
    """
    def __hash__(self):
        if self._hash is None:
            self._hash = hash(frozenset(viewitems(self)))
        return self._hash



"""
Python 2/3 compatibility helpers.
"""

from sys import version_info

PY2 = version_info[0] == 2
if PY2:
    assert version_info[1] > 6, 'Python >= 2.7 required'

if PY2:
    iteritems = lambda x: x.iteritems()
    viewitems = lambda x: x.viewitems()
else:
    iteritems = lambda x: iter(x.items())
    viewitems = lambda x: x.items()

iteritems.__doc__ = 'Python 2/3 compatible iteritems'
viewitems.__doc__ = 'Python 2/3 compatible viewitems'



"""
Utilities for working with one-to-one relations.
"""

from .compat import PY2, iteritems
from collections import Iterator


def pairs(*map_or_it, **kw):
    """
    Yields the pairs provided. Signature matches dict's.
    Accepts zero or one positional argument which it first tries iterating over
    as a mapping, and if that fails, falls back to iterating over as
    a sequence, yielding items two at a time.
    Mappings may also be passed as keyword arguments, which will be yielded
    after any passed via positional argument.
    """
    if map_or_it:
        l = len(map_or_it)
        if l != 1:
            raise TypeError('expected at most 1 argument, got %d' % l)
        map_or_it = map_or_it[0]
        try:
            it = iteritems(map_or_it)   # mapping?
        except AttributeError:          #  no
            for (k, v) in map_or_it:    #    -> treat as sequence
                yield (k, v)
        else:                           #  yes
            for (k, v) in it:           #    -> treat as mapping
                yield (k, v)
    for (k, v) in iteritems(kw):
        yield (k, v)


class inverted(Iterator):
    """
    An iterator analogous to the :func:`reversed` built-in.
    Useful for inverting a mapping.
    """
    def __init__(self, data):
        self._data = data

    def __iter__(self):
        """
        First try to call ``__inverted__`` on the wrapped object
        and return the result if the call succeeds
        (i.e. delegate to the object if it supports inverting natively).
        This complements :attr:`bidict.BidirectionalMapping.__inverted__`.

        If the call fails, fall back on calling our own ``__next__`` method.
        """
        try:
            it = self._data.__inverted__
        except AttributeError:
            it = self.__next__
        return it()

    def __next__(self):
        """
        Yields the inverse of each pair yielded by calling :attr:`bidict.pairs`
        on the wrapped object.
        """
        for (k, v) in pairs(self._data):
            yield (v, k)

    # compat
    if PY2:
        next = __next__



from ._bidict import bidict

class collapsingbidict(bidict):
    """
    A mutable bidict which does not throw a :class:`bidict.CollapseException`
    when an update would cause two existing mappings to collapse,
    but rather allows the update to succeed
    with no explicit sign that two mappings just collapsed into one::

        >>> b = collapsingbidict({0: 'zero', 1: 'one'})
        >>> b[0] = 'one'
        >>> b
        collapsingbidict({0: 'one'})

    """
    _put = bidict.forceput



from .compat import PY2, iteritems, viewitems
from .util import pairs
from collections import Mapping


class BidirectionalMapping(Mapping):
    """
    Mutable and immutable bidict types extend this class,
    which implements all the shared logic.
    Users typically won't need to touch this.
    """
    def __init__(self, *args, **kw):
        self._fwd = {}
        self._bwd = {}
        for (k, v) in pairs(*args, **kw):
            self._put(k, v)
        inv = object.__new__(self.__class__)
        inv._fwd = self._bwd
        inv._bwd = self._fwd
        inv._inv = self
        self._inv = inv
        self._hash = None

    def __repr__(self):
        return '%s(%r)' % (self.__class__.__name__, self._fwd)

    __str__ = __repr__

    def __eq__(self, other):
        return self._fwd == other

    def __ne__(self, other):
        return self._fwd != other

    def __invert__(self):
        """
        Called when the unary inverse operator (~) is applied.
        """
        return self._inv

    inv = property(__invert__, doc='Property providing access to the inverse '
                   'bidict. Can be chained as in: ``B.inv.inv is B``')

    def __inverted__(self):
        return iteritems(self._bwd)

    @staticmethod
    def _fwd_slice(slice):
        """
        Raises :class:`TypeError` if the given slice does not have either only
        its start or only its stop set to a non-None value.

        Returns True if only its start is not None and False if only its stop
        is not None.
        """
        if slice.step is not None:
            raise TypeError('Slice may not specify step')
        none_start = slice.start is None
        none_stop = slice.stop is None
        if none_start == none_stop:
            raise TypeError('Exactly one of slice start or stop must be None '
                            'and the other must not be')
        return not none_start

    def __getitem__(self, keyorslice):
        """
        Provides a __getitem__ implementation
        which accepts a slice (e.g. ``b[:val]``)
        to allow referencing an inverse mapping.
        A non-slice value (e.g. ``b[key]``)
        is considered a reference to a forward mapping.
        """
        if isinstance(keyorslice, slice):
            # forward lookup (by key): b[key:]
            if self._fwd_slice(keyorslice):
                return self._fwd[keyorslice.start]
            else:  # inverse lookup (by val): b[:val]
                return self._bwd[keyorslice.stop]
        else:  # keyorslice is a key: b[key]
            return self._fwd[keyorslice]

    def _put(self, key, val):
        try:
            oldval = self._fwd[key]
        except KeyError:
            oldval = _sentinel
        try:
            oldkey = self._bwd[val]
        except KeyError:
            oldkey = _sentinel

        if oldval is not _sentinel and oldkey is not _sentinel:
            if key == oldkey and val == oldval:
                return
            raise CollapseException((key, oldval), (oldkey, val))
        elif oldval is not _sentinel:
            del self._bwd[oldval]
        elif oldkey is not _sentinel:
            del self._fwd[oldkey]

        self._fwd[key] = val
        self._bwd[val] = key

    get = lambda self, k, *args: self._fwd.get(k, *args)
    copy = lambda self: self.__class__(self._fwd)
    get.__doc__ = dict.get.__doc__
    copy.__doc__ = dict.copy.__doc__
    __len__ = lambda self: len(self._fwd)
    __iter__ = lambda self: iter(self._fwd)
    __contains__ = lambda self, x: x in self._fwd
    __len__.__doc__ = dict.__len__.__doc__
    __iter__.__doc__ = dict.__iter__.__doc__
    __contains__.__doc__ = dict.__contains__.__doc__
    keys = lambda self: self._fwd.keys()
    items = lambda self: self._fwd.items()
    keys.__doc__ = dict.keys.__doc__
    items.__doc__ = dict.items.__doc__
    values = lambda self: self._bwd.keys()
    values.__doc__ = \
        "D.values() -> a set-like object providing a view on D's values. " \
        'Note that because values of a BidirectionalMapping are also keys, ' \
        'this returns a ``dict_keys`` object rather than a ``dict_values`` ' \
        'object.'
    if PY2:
        iterkeys = lambda self: self._fwd.iterkeys()
        viewkeys = lambda self: self._fwd.viewkeys()
        iteritems = lambda self: self._fwd.iteritems()
        viewitems = lambda self: self._fwd.viewitems()
        itervalues = lambda self: self._bwd.iterkeys()
        viewvalues = lambda self: self._bwd.viewkeys()
        iterkeys.__doc__ = dict.iterkeys.__doc__
        viewkeys.__doc__ = dict.viewkeys.__doc__
        iteritems.__doc__ = dict.iteritems.__doc__
        viewitems.__doc__ = dict.viewitems.__doc__
        itervalues.__doc__ = dict.itervalues.__doc__
        viewvalues.__doc__ = values.__doc__.replace('values()', 'viewvalues()')
        values.__doc__ = dict.values.__doc__


class CollapseException(Exception):
    """
    Raised when an attempt is made to insert a new mapping into a bidict that
    would collapse two existing mappings.
    """


_sentinel = object()



from ._common import BidirectionalMapping, _sentinel
from .util import pairs
from collections import MutableMapping


class bidict(BidirectionalMapping, MutableMapping):
    """
    Mutable bidirectional map type.
    """
    def _del(self, key):
        val = self._fwd[key]
        del self._fwd[key]
        del self._bwd[val]

    def __delitem__(self, keyorslice):
        if isinstance(keyorslice, slice):
            # delete by key: del b[key:]
            if self._fwd_slice(keyorslice):
                self._del(keyorslice.start)
            else:  # delete by value: del b[:val]
                self._del(self._bwd[keyorslice.stop])
        else:  # keyorslice is a key: del b[key]
            self._del(keyorslice)

    def __setitem__(self, keyorslice, keyorval):
        if isinstance(keyorslice, slice):
            # keyorslice.start is key, keyorval is val: b[key:] = val
            if self._fwd_slice(keyorslice):
                self._put(keyorslice.start, keyorval)
            else:  # keyorval is key, keyorslice.stop is val: b[:val] = key
                self._put(keyorval, keyorslice.stop)
        else:  # keyorslice is a key, keyorval is a val: b[key] = val
            self._put(keyorslice, keyorval)

    def put(self, key, val):
        """
        Alternative to using the setitem syntax to insert a mapping.
        """
        self._put(key, val)

    def forceput(self, key, val):
        """
        Like :attr:`bidict.bidict.put` but silently removes any existing
        mapping that would otherwise cause a :class:`bidict.CollapseException`
        before inserting the given mapping::

            >>> b = bidict({0: 'zero', 1: 'one'})
            >>> b.put(0, 'one')  # doctest: +IGNORE_EXCEPTION_DETAIL
            Traceback (most recent call last):
            ...
            CollapseException: ((0, 'zero'), (1, 'one'))
            >>> b.forceput(0, 'one')
            >>> b
            bidict({0: 'one'})

        """
        oldval = self._fwd.get(key, _sentinel)
        oldkey = self._bwd.get(val, _sentinel)
        if oldval is not _sentinel:
            del self._bwd[oldval]
        if oldkey is not _sentinel:
            del self._fwd[oldkey]
        self._fwd[key] = val
        self._bwd[val] = key

    def clear(self):
        self._fwd.clear()
        self._bwd.clear()

    def invert(self):
        self._fwd, self._bwd = self._bwd, self._fwd
        self._inv._fwd, self._inv._bwd = self._inv._bwd, self._inv._fwd

    def pop(self, key, *args):
        val = self._fwd.pop(key, *args)
        del self._bwd[val]
        return val

    def popitem(self):
        if not self._fwd:
            raise KeyError('popitem(): %s is empty' % self.__class__.__name__)
        key, val = self._fwd.popitem()
        del self._bwd[val]
        return key, val

    def setdefault(self, key, default=None):
        val = self._fwd.setdefault(key, default)
        self._bwd[val] = key
        return val

    def update(self, *args, **kw):
        for k, v in pairs(*args, **kw):
            self[k] = v



# -*- coding: utf-8 -*-

"""
Efficient, Pythonic bidirectional map implementation
and related functionality.

.. :copyright: (c) 2015 Joshua Bronson.
.. :license: ISCL. See LICENSE for details.

"""

from ._common import BidirectionalMapping, CollapseException
from ._bidict import bidict
from ._collapsing import collapsingbidict
from ._frozen import frozenbidict
from ._named import namedbidict
from .util import pairs, inverted

__all__ = (
    'BidirectionalMapping',
    'CollapseException',
    'bidict',
    'collapsingbidict',
    'frozenbidict',
    'namedbidict',
    'pairs',
    'inverted',
)

try:
    from pkg_resources import resource_string
    __version__ = resource_string(__name__, 'VERSION').decode('ascii').strip()
except Exception as e:
    from warnings import warn
    warn('Failed to read/set version: %r' % e)



from ._bidict import bidict
from re import compile as compile_re

_LEGALNAMEPAT = '^[a-zA-Z][a-zA-Z0-9_]*$'
_LEGALNAMERE = compile_re(_LEGALNAMEPAT)

def _empty_namedbidict(mapname, fwdname, invname):
    """
    Create an empty instance of a custom bidict.
    Used to make :func:`bidict.namedbidict` instances picklable.
    """
    return namedbidict(mapname, fwdname, invname)()

def namedbidict(mapname, fwdname, invname, base_type=bidict):
    """
    Allows creating custom named bidict types,
    analagous to :func:`collections.namedtuple`.
    """
    for name in mapname, fwdname, invname:
        if not _LEGALNAMERE.match(name):
            raise ValueError('"%s" does not match pattern %s' %
                             (name, _LEGALNAMEPAT))

    for_fwd = invname + '_for'
    for_inv = fwdname + '_for'
    __dict__ = {for_fwd: property(lambda self: self),
                for_inv: base_type.inv}

    custombidict = type(mapname, (base_type,), __dict__)

    # support pickling
    custombidict.__reduce__ = lambda self: \
        (_empty_namedbidict, (mapname, fwdname, invname), self.__dict__)

    return custombidict
